/*
	Copyright (C) 2004-2005 Cory Nelson

	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any damages
	arising from the use of this software.

	Permission is granted to anyone to use this software for any purpose,
	including commercial applications, and to alter it and redistribute it
	freely, subject to the following restrictions:

	1. The origin of this software must not be misrepresented; you must not
		claim that you wrote the original software. If you use this software
		in a product, an acknowledgment in the product documentation would be
		appreciated but is not required.
	2. Altered source versions must be plainly marked as such, and must not be
		misrepresented as being the original software.
	3. This notice may not be removed or altered from any source distribution.
	
	CVS Info :
		$Author: phrostbyte $
		$Date: 2005/04/20 03:30:47 $
		$Revision: 1.16 $
*/

#include "stdafx.h"
#include "resource.h"
using namespace std;

static const LPCTSTR IPFILTER_NAME=_T("IpFilterDriver");
static const LPCTSTR IPFILTER_PATH=_T("System32\\Drivers\\IpFltDrv.sys");
static const LPCTSTR PGFILTER_NAME=_T("pgfilter");
static const LPCTSTR PGFILTER_PATH=_T("pgfilter.sys");

PgFilter::PgFilter() : block(true),blockhttp(true),runthread(true) {
	if(this->isrunning()) throw runtime_error("driver already running");

	this->ipfltdrv.Load(IPFILTER_NAME, IPFILTER_PATH);
	this->ipfltdrv.removable=false;
	this->ipfltdrv.Start();

	this->filter.Load(PGFILTER_NAME, (path::base_dir()/PGFILTER_PATH).file_str().c_str());
	this->filter.removable=true;
	this->filter.Start();

	this->thread=boost::shared_ptr<::thread>(new ::thread(boost::bind(&PgFilter::notifythread, this)));
}

PgFilter::~PgFilter() {
	this->runthread=false;

	{
		try {
			this->filter.Close();
		}
		catch(exception &ex) {
			ExceptionBox(NULL, ex, __FILE__, __LINE__);
		}
	}

	{
		try {
			this->ipfltdrv.Close();
		}
		catch(exception &ex) {
			ExceptionBox(NULL, ex, __FILE__, __LINE__);
		}
	}

	this->thread->join();
}

void PgFilter::SetRanges(const p2p::list &ranges) {
	boost::scoped_array<PGIPRANGE> r(new PGIPRANGE[ranges.size()]);

	map<wstring,LPCTSTR> names;
	for(p2p::list::const_iterator iter=ranges.begin(); iter!=ranges.end(); iter++)
		names[iter->name]=NULL;

	size_t nameslen=names.size();
	for(map<wstring,LPCTSTR>::const_iterator iter=names.begin(); iter!=names.end(); iter++)
		nameslen+=WCHAR_TSTRING(iter->first).length();

	boost::shared_array<TCHAR> namesbuf(new TCHAR[nameslen]);
	TCHAR *bufiter=namesbuf.get();

	for(map<wstring,LPCTSTR>::iterator iter=names.begin(); iter!=names.end(); iter++) {
		tstring str=WCHAR_TSTRING(iter->first);
		LPCTSTR ptr=str.c_str();
		size_t len=str.length()+1;
		copy(ptr, ptr+len, bufiter);

		iter->second=bufiter;
		bufiter+=len;
	}

	unsigned int blockedcount=(unsigned int)ranges.size();

	size_t i=0;
	for(p2p::list::const_iterator iter=ranges.begin(); iter!=ranges.end(); iter++) {
		r[i].name=names[iter->name];
		r[i].start=iter->start.ipl;
		r[i++].end=iter->end.ipl;

		blockedcount+=iter->end.ipl - iter->start.ipl;
	}
	
	{
		mutex::scoped_lock lock(this->mutex);

		if(this->block) {
			int data=0;
			this->filter.Write(IOCTL_PEERGUARDIAN_HOOK, &data, sizeof(data));
		}

		this->ranges=namesbuf;
		this->blockedcount=blockedcount;

		this->filter.Write(IOCTL_PEERGUARDIAN_SETRANGES, r.get(), (DWORD)(ranges.size()*sizeof(PGIPRANGE)));

		if(this->block) {
			int data=1;
			this->filter.Write(IOCTL_PEERGUARDIAN_HOOK, &data, sizeof(data));
		}
	}
}

void PgFilter::SetBlock(bool block) {
	mutex::scoped_lock lock(this->mutex);

	if(block!=this->block) {
		this->block=block;

      int data=block?1:0;
		this->filter.Write(IOCTL_PEERGUARDIAN_HOOK, &data, sizeof(data));
	}
}

void PgFilter::SetBlockHttp(bool block) {
	mutex::scoped_lock lock(this->mutex);

	if(block!=this->blockhttp) {
		this->blockhttp=block;

		int data=block?1:0;
		this->filter.Write(IOCTL_PEERGUARDIAN_HTTP, &data, sizeof(data));
	}
}

void PgFilter::notifythread() {
	OVERLAPPED overlapped;
	PGNOTIFICATION notification;

	HANDLE event=CreateEvent(NULL, TRUE, FALSE, NULL);

	while(this->runthread) {
		ZeroMemory(&overlapped, sizeof(overlapped));
		ZeroMemory(&notification, sizeof(notification));

		overlapped.hEvent=event;

		if(!this->filter.Read(IOCTL_PEERGUARDIAN_GETNOTIFICATION, &notification, sizeof(notification), &overlapped)) {
			DWORD err=GetLastError();
			if(err==ERROR_OPERATION_ABORTED) break;
			else if(err!=ERROR_IO_PENDING) {
				LPTSTR str;

				FormatMessage(
					FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
					NULL,
					err,
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					(LPTSTR)&str,
					0, NULL);

				MessageBox(NULL, str, LoadString(IDS_EXCEPTION), MB_ICONERROR|MB_OK);
				LocalFree(str);
			}
		}

		while(this->runthread && WaitForSingleObject(event, 500)==WAIT_TIMEOUT);
		if(this->runthread) {
			if(notification.source!=0 && notification.destination!=0 && !this->OnAction.empty()) {
				Action a;

				if(!IsBadReadPtr(notification.name, 1)) {
					try {
						a.Name=(notification.name==NULL)?_T("n/a"):notification.name;
					}
					catch(...) {
						a.Name=_T("n/a");
					}
				}
				else a.Name=_T("n/a");

				a.SourceIp=notification.source;
				a.SourcePort=notification.sourceport;
				a.DestIp=notification.destination;
				a.DestPort=notification.destinationport;
				a.Blocked=(notification.blocked==1);
				a.Protocol=notification.protocol;

				try {
					this->OnAction(a);
				}
				catch(exception &ex) {
					ExceptionBox(NULL, ex, __FILE__, __LINE__);
				}
			}

			ResetEvent(event);
		}
	}

	CloseHandle(event);
}

bool PgFilter::isrunning() {
	SC_HANDLE manager=OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if(!manager) throw win32_error("OpenSCManager");

	SC_HANDLE service=OpenService(manager, PGFILTER_NAME, SERVICE_ALL_ACCESS);
	if(!service) {
		CloseServiceHandle(manager);
		return false;
	}

	SERVICE_STATUS status={0};
	if(QueryServiceStatus(service, &status) && status.dwCurrentState!=SERVICE_STOPPED) {
		/*if(!ControlService(service, SERVICE_CONTROL_STOP, &status)) {
			CloseServiceHandle(service);
			CloseServiceHandle(manager);
			return true;
		}*/

		CloseServiceHandle(service);
		CloseServiceHandle(manager);
		return true;
	}

	bool ret=(DeleteService(service)==FALSE);
	if(ret) ret=(GetLastError()!=ERROR_SERVICE_MARKED_FOR_DELETE);

	CloseServiceHandle(service);
	CloseServiceHandle(manager);

	return ret;
}
